@@ -7,12 +7,14 @@
 
 //! UPB FFI wrapper code for use by Rust Protobuf.
 
-use crate::__internal::{Enum, Private, SealedInternal};
+use crate::__internal::{Enum, MatcherEq, Private, SealedInternal};
 use crate::{
-    IntoProxied, Map, MapIter, MapMut, MapView, Mut, ProtoBytes, ProtoStr, ProtoString, Proxied,
-    ProxiedInMapValue, ProxiedInRepeated, Repeated, RepeatedMut, RepeatedView, View,
+    AsMut, AsView, Clear, ClearAndParse, CopyFrom, IntoProxied, Map, MapIter, MapMut, MapView,
+    MergeFrom, Message, Mut, ParseError, ProtoBytes, ProtoStr, ProtoString, Proxied,
+    ProxiedInMapValue, ProxiedInRepeated, Repeated, RepeatedMut, RepeatedView, TakeFrom, View,
 };
-use core::fmt::Debug;
+use std::fmt::Debug;
+use std::marker::PhantomData;
 use std::mem::{size_of, ManuallyDrop, MaybeUninit};
 use std::ptr::{self, NonNull};
 use std::slice;
@@ -33,6 +35,18 @@ pub type RawRepeatedField = upb::RawArray;
 pub type RawMap = upb::RawMap;
 pub type PtrAndLen = upb::StringView;
 
+// This struct represents a raw minitable pointer. We need it to be Send and Sync so that we can
+// store it in a static OnceLock for lazy initialization of minitables. It should not be used for
+// any other purpose.
+pub struct MiniTablePtr(pub *mut upb_MiniTable);
+unsafe impl Send for MiniTablePtr {}
+unsafe impl Sync for MiniTablePtr {}
+
+// Same as above, but for enum minitables.
+pub struct MiniTableEnumPtr(pub *const upb_MiniTableEnum);
+unsafe impl Send for MiniTableEnumPtr {}
+unsafe impl Sync for MiniTableEnumPtr {}
+
 impl From<&ProtoStr> for PtrAndLen {
     fn from(s: &ProtoStr) -> Self {
         let bytes = s.as_bytes();
@@ -61,6 +75,12 @@ impl ScratchSpace {
     }
 }
 
+thread_local! {
+    // We need to avoid dropping this Arena, because we use it to build mini tables that
+    // effectively have 'static lifetimes.
+    pub static THREAD_LOCAL_ARENA: ManuallyDrop<Arena> = ManuallyDrop::new(Arena::new());
+}
+
 #[doc(hidden)]
 pub type SerializedData = upb::OwnedArenaBox<[u8]>;
 
@@ -72,12 +92,51 @@ impl IntoProxied<ProtoBytes> for SerializedData {
     }
 }
 
-/// The raw contents of every generated message.
 #[derive(Debug)]
 #[doc(hidden)]
-pub struct MessageInner {
-    pub msg: RawMessage,
-    pub arena: Arena,
+pub struct OwnedMessageInner<T> {
+    ptr: MessagePtr<T>,
+    arena: Arena,
+}
+
+impl<T: Message + AssociatedMiniTable> Default for OwnedMessageInner<T> {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
+impl<T: Message + AssociatedMiniTable> OwnedMessageInner<T> {
+    pub fn new() -> Self {
+        let arena = Arena::new();
+        let ptr = MessagePtr::new(&arena).expect("alloc should never fail");
+        Self { ptr, arena }
+    }
+
+    /// # Safety
+    /// - The underlying pointer must of type `T` and be allocated on `arena`.
+    pub unsafe fn wrap_raw(raw: RawMessage, arena: Arena) -> Self {
+        // SAFETY:
+        // - Caller guaranteed `raw` is valid and of type `T`
+        let ptr = unsafe { MessagePtr::wrap(raw) };
+        OwnedMessageInner { ptr, arena }
+    }
+
+    pub fn ptr_mut(&mut self) -> MessagePtr<T> {
+        self.ptr
+    }
+
+    pub fn ptr(&self) -> MessagePtr<T> {
+        self.ptr
+    }
+
+    pub fn raw(&self) -> RawMessage {
+        self.ptr.raw()
+    }
+
+    #[allow(clippy::needless_pass_by_ref_mut)] // Sound access requires mutable access.
+    pub fn arena(&mut self) -> &Arena {
+        &self.arena
+    }
 }
 
 /// Mutators that point to their original message use this to do so.
@@ -90,11 +149,11 @@ pub struct MessageInner {
 ///   place any restriction on the layout of generated messages and their
 ///   mutators. This makes a vtable-based mutator three pointers, which can no
 ///   longer be returned in registers on most platforms.
-/// - Store one pointer here, `&'msg MessageInner`, where `MessageInner` stores
+/// - Store one pointer here, `&'msg OwnedMessageInner`, where `OwnedMessageInner` stores
 ///   a `RawMessage` and an `Arena`. This would require all generated messages
-///   to store `MessageInner`, and since their mutators need to be able to
+///   to store `OwnedMessageInner`, and since their mutators need to be able to
 ///   generate `BytesMut`, would also require `BytesMut` to store a `&'msg
-///   MessageInner` since they can't store an owned `Arena`.
+///   OwnedMessageInner` since they can't store an owned `Arena`.
 ///
 /// Note: even though this type is `Copy`, it should only be copied by
 /// protobuf internals that can maintain mutation invariants:
@@ -105,26 +164,54 @@ pub struct MessageInner {
 ///   must be different fields, and not be in the same oneof. As such, a `Mut`
 ///   cannot be `Clone` but *can* reborrow itself with `.as_mut()`, which
 ///   converts `&'b mut Mut<'a, T>` to `Mut<'b, T>`.
-#[derive(Clone, Copy, Debug)]
+#[derive(Debug)]
 #[doc(hidden)]
-pub struct MutatorMessageRef<'msg> {
-    msg: RawMessage,
+pub struct MessageMutInner<'msg, T> {
+    ptr: MessagePtr<T>,
     arena: &'msg Arena,
 }
 
-impl<'msg> MutatorMessageRef<'msg> {
-    #[doc(hidden)]
+impl<'msg, T: Message + AssociatedMiniTable> Clone for MessageMutInner<'msg, T> {
+    fn clone(&self) -> Self {
+        *self
+    }
+}
+impl<'msg, T: Message + AssociatedMiniTable> Copy for MessageMutInner<'msg, T> {}
+
+impl<'msg, T: Message + AssociatedMiniTable> MessageMutInner<'msg, T> {
+    /// # Safety
+    /// - `msg` must be a valid `RawMessage`
+    /// - `arena` must hold the memory for `msg`
+    pub unsafe fn wrap_raw(raw: RawMessage, arena: &'msg Arena) -> Self {
+        // SAFETY:
+        // - Caller guaranteed `raw` is valid and of type `T`
+        let ptr = unsafe { MessagePtr::wrap(raw) };
+        MessageMutInner { ptr, arena }
+    }
+
     #[allow(clippy::needless_pass_by_ref_mut)] // Sound construction requires mutable access.
-    pub fn new(msg: &'msg mut MessageInner) -> Self {
-        MutatorMessageRef { msg: msg.msg, arena: &msg.arena }
+    pub fn mut_of_owned(msg: &'msg mut OwnedMessageInner<T>) -> Self {
+        MessageMutInner { ptr: msg.ptr, arena: &msg.arena }
     }
 
-    pub fn from_parent(parent_msg: MutatorMessageRef<'msg>, message_field_ptr: RawMessage) -> Self {
-        MutatorMessageRef { msg: message_field_ptr, arena: parent_msg.arena }
+    pub fn from_parent<ParentT>(
+        parent_msg: MessageMutInner<'msg, ParentT>,
+        raw_field_ptr: RawMessage,
+    ) -> Self {
+        let ptr = unsafe { MessagePtr::wrap(raw_field_ptr) };
+        MessageMutInner { ptr, arena: parent_msg.arena }
     }
 
-    pub fn msg(&self) -> RawMessage {
-        self.msg
+    pub fn ptr_mut(&mut self) -> MessagePtr<T> {
+        self.ptr
+    }
+
+    pub fn ptr(&self) -> MessagePtr<T> {
+        self.ptr
+    }
+
+    pub fn raw(&self) -> RawMessage {
+        self.ptr.raw()
     }
 
     pub fn arena(&self) -> &Arena {
@@ -132,6 +219,57 @@ impl<'msg> MutatorMessageRef<'msg> {
     }
 }
 
+#[derive(Debug)]
+#[doc(hidden)]
+pub struct MessageViewInner<'msg, T> {
+    ptr: MessagePtr<T>,
+    _phantom: PhantomData<&'msg ()>,
+}
+
+impl<'msg, T: Message + AssociatedMiniTable> Clone for MessageViewInner<'msg, T> {
+    fn clone(&self) -> Self {
+        *self
+    }
+}
+impl<'msg, T: Message + AssociatedMiniTable> Copy for MessageViewInner<'msg, T> {}
+
+impl<'msg, T: Message + AssociatedMiniTable> MessageViewInner<'msg, T> {
+    /// # Safety
+    /// - The underlying pointer must live as long as `'msg`.
+    pub unsafe fn wrap(ptr: MessagePtr<T>) -> Self {
+        // SAFETY:
+        // - Caller guaranteed `raw` is valid
+        MessageViewInner { ptr, _phantom: PhantomData }
+    }
+
+    /// # Safety
+    /// - The underlying pointer must of type `T` and live as long as `'msg`.
+    pub unsafe fn wrap_raw(raw: RawMessage) -> Self {
+        // SAFETY:
+        // - Caller guaranteed `raw` is valid and of type `T`
+        let ptr = unsafe { MessagePtr::wrap(raw) };
+        MessageViewInner { ptr, _phantom: PhantomData }
+    }
+
+    #[allow(clippy::needless_pass_by_ref_mut)] // Sound construction requires mutable access.
+    pub fn view_of_owned(owned: &'msg OwnedMessageInner<T>) -> Self {
+        MessageViewInner { ptr: owned.ptr, _phantom: PhantomData }
+    }
+
+    #[allow(clippy::needless_pass_by_ref_mut)] // Sound construction requires mutable access.
+    pub fn view_of_mut(msg_mut: MessageMutInner<'msg, T>) -> Self {
+        MessageViewInner { ptr: msg_mut.ptr, _phantom: PhantomData }
+    }
+
+    pub fn ptr(&self) -> MessagePtr<T> {
+        self.ptr
+    }
+
+    pub fn raw(&self) -> RawMessage {
+        self.ptr.raw()
+    }
+}
+
 /// Kernel-specific owned `string` and `bytes` field type.
 #[doc(hidden)]
 pub struct InnerProtoString(OwnedArenaBox<[u8]>);
@@ -347,12 +485,128 @@ macro_rules! impl_repeated_bytes {
     }
 }
 
+unsafe impl<T> ProxiedInRepeated for T
+where
+    T: Message + UpbTypeConversions + AssociatedMiniTable,
+{
+    fn repeated_new(_private: Private) -> Repeated<Self> {
+        let arena = Arena::new();
+        Repeated::from_inner(Private, unsafe {
+            InnerRepeated::from_raw_parts(upb_Array_New(arena.raw(), CType::Message), arena)
+        })
+    }
+
+    unsafe fn repeated_free(_private: Private, _repeated: &mut Repeated<Self>) {
+        // No-op: the memory will be dropped by the arena.
+    }
+
+    fn repeated_len(repeated: View<Repeated<Self>>) -> usize {
+        // SAFETY: `repeated.as_raw()` is a valid `upb_Array*`.
+        unsafe { upb_Array_Size(repeated.as_raw(Private)) }
+    }
+
+    fn repeated_push(mut repeated: Mut<Repeated<Self>>, val: impl IntoProxied<Self>) {
+        // SAFETY:
+        // - `repeated.as_raw()` is a valid `upb_Array*`.
+        // - `msg_ptr` is a valid `upb_Message*`.
+        unsafe {
+            upb_Array_Append(
+                repeated.as_raw(Private),
+                <Self as UpbTypeConversions>::into_message_value_fuse_if_required(
+                    repeated.raw_arena(Private),
+                    val.into_proxied(Private),
+                ),
+                repeated.raw_arena(Private),
+            );
+        };
+    }
+
+    fn repeated_clear(mut repeated: Mut<Repeated<Self>>) {
+        // SAFETY:
+        // - `repeated.as_raw()` is a valid `upb_Array*`.
+        unsafe { upb_Array_Resize(repeated.as_raw(Private), 0, repeated.raw_arena(Private)) };
+    }
+
+    unsafe fn repeated_get_unchecked<'a>(
+        repeated: View<'a, Repeated<Self>>,
+        index: usize,
+    ) -> View<'a, Self> {
+        // SAFETY:
+        // - `repeated.as_raw()` is a valid `const upb_Array*`.
+        // - `index < len(repeated)` is promised by the caller.
+        let val = unsafe { upb_Array_Get(repeated.as_raw(Private), index) };
+        // SAFETY:
+        // - `val` has the correct variant for Self.
+        // - `val` is valid for `'a` lifetime.
+        unsafe { Self::from_message_value(val) }
+    }
+
+    unsafe fn repeated_get_mut_unchecked<'a>(
+        mut repeated: Mut<'a, Repeated<Self>>,
+        index: usize,
+    ) -> Mut<'a, Self> {
+        // SAFETY:
+        // - `repeated.as_raw()` is a valid `upb_Array*`.
+        // - `repeated` is a an array of message-valued elements.
+        // - `index < len(repeated)` is promised by the caller.
+        let val = unsafe { upb_Array_GetMutable(repeated.as_raw(Private), index) };
+        // SAFETY:
+        // - `val` is the correct variant for `Self`.
+        // - `val` is valid for `'a` lifetime.
+        unsafe { Self::from_message_mut(val, repeated.arena(Private)) }
+    }
+
+    unsafe fn repeated_set_unchecked(
+        mut repeated: Mut<Repeated<Self>>,
+        index: usize,
+        val: impl IntoProxied<Self>,
+    ) {
+        unsafe {
+            upb_Array_Set(
+                repeated.as_raw(Private),
+                index,
+                <Self as UpbTypeConversions>::into_message_value_fuse_if_required(
+                    repeated.raw_arena(Private),
+                    val.into_proxied(Private),
+                ),
+            )
+        }
+    }
+
+    fn repeated_copy_from(src: View<Repeated<Self>>, dest: Mut<Repeated<Self>>) {
+        // SAFETY:
+        // - Elements of `src` and `dest` have message minitable `Self::mini_table()`.
+        unsafe {
+            repeated_message_copy_from(src, dest, <Self as AssociatedMiniTable>::mini_table());
+        }
+    }
+
+    fn repeated_reserve(mut repeated: Mut<Repeated<Self>>, additional: usize) {
+        // SAFETY:
+        // - `repeated.as_raw()` is a valid `upb_Array*`.
+        unsafe {
+            let size = upb_Array_Size(repeated.as_raw(Private));
+            upb_Array_Reserve(
+                repeated.as_raw(Private),
+                size + additional,
+                repeated.raw_arena(Private),
+            );
+        }
+    }
+}
+
 impl<'msg, T> RepeatedMut<'msg, T> {
     // Returns a `RawArena` which is live for at least `'msg`
     #[doc(hidden)]
     pub fn raw_arena(&mut self, _private: Private) -> RawArena {
         self.inner.arena.raw()
     }
+
+    // Returns an `Arena` which is live for at least `'msg`
+    #[doc(hidden)]
+    pub fn arena(&self, _private: Private) -> &'msg Arena {
+        self.inner.arena
+    }
 }
 
 impl_repeated_primitives!(
@@ -505,6 +759,12 @@ impl<'msg, K: ?Sized, V: ?Sized> MapMut<'msg, K, V> {
     pub fn raw_arena(&mut self, _private: Private) -> RawArena {
         self.inner.arena.raw()
     }
+
+    // Returns an `Arena` which is live for at least `'msg`
+    #[doc(hidden)]
+    pub fn arena(&self, _private: Private) -> &'msg Arena {
+        self.inner.arena
+    }
 }
 
 #[derive(Debug)]
@@ -565,9 +825,20 @@ pub trait UpbTypeConversions: Proxied {
     ) -> upb_MessageValue;
 
     /// # Safety
-    /// - `msg` must be the correct variant for `Self`.
-    /// - `msg` pointers must point to memory valid for `'msg` lifetime.
-    unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, Self>;
+    /// - `msg_val` must be the correct variant for `Self`.
+    /// - `msg_val` pointers must point to memory valid for `'msg` lifetime.
+    unsafe fn from_message_value<'msg>(msg_val: upb_MessageValue) -> View<'msg, Self>;
+
+    /// # Safety
+    /// - `raw` must be the correct variant for `Self`.
+    /// - `raw` pointers must point to memory valid for `'msg` lifetime.
+    #[allow(unused_variables)]
+    unsafe fn from_message_mut<'msg>(raw: RawMessage, arena: &'msg Arena) -> Mut<'msg, Self>
+    where
+        Self: Message,
+    {
+        panic!("mut_from_message_value is only implemented for messages.")
+    }
 }
 
 macro_rules! impl_upb_type_conversions_for_scalars {
@@ -753,6 +1024,23 @@ where
         Some(unsafe { <Self as UpbTypeConversions>::from_message_value(val.assume_init()) })
     }
 
+    fn map_get_mut<'a>(mut map: MapMut<'a, Key, Self>, key: View<'_, Key>) -> Option<Mut<'a, Self>>
+    where
+        Self: Message,
+    {
+        // SAFETY: The map is valid as promised by the caller.
+        let val = unsafe {
+            upb_Map_GetMutable(
+                map.as_raw(Private),
+                <Key as UpbTypeConversions>::to_message_value(key),
+            )
+        };
+        // SAFETY: The lifetime of the MapMut is guaranteed to outlive the returned Mut.
+        NonNull::new(val).map(|msg| unsafe {
+            <Self as UpbTypeConversions>::from_message_mut(msg, map.arena(Private))
+        })
+    }
+
     fn map_remove(mut map: MapMut<Key, Self>, key: View<'_, Key>) -> bool {
         unsafe {
             upb_Map_Delete(
@@ -808,3 +1096,151 @@ pub unsafe fn upb_Map_InsertAndReturnIfInserted(
         upb::MapInsertStatus::OutOfMemory => panic!("map arena is out of memory"),
     }
 }
+
+/// Internal-only trait to support blanket impls that need const access to raw messages
+/// on codegen. Should never be used by application code.
+#[doc(hidden)]
+pub unsafe trait UpbGetMessagePtr: SealedInternal {
+    type Msg: AssociatedMiniTable + Message;
+
+    fn get_ptr(&self, _private: Private) -> MessagePtr<Self::Msg>;
+
+    // TODO: Remove.
+    fn get_raw_message(&self, _private: Private) -> RawMessage {
+        self.get_ptr(Private).raw()
+    }
+}
+
+/// Internal-only trait to support blanket impls that need mutable access to raw messages
+/// on codegen. Must not be implemented on View proxies. Should never be used by application code.
+#[doc(hidden)]
+pub unsafe trait UpbGetMessagePtrMut: SealedInternal {
+    type Msg: AssociatedMiniTable + Message;
+
+    fn get_ptr_mut(&mut self, _private: Private) -> MessagePtr<Self::Msg>;
+
+    // TODO: remove.
+    fn get_raw_message_mut(&mut self, _private: Private) -> RawMessage {
+        self.get_ptr_mut(Private).raw()
+    }
+}
+
+/// Internal-only trait to support blanket impls that need const access to raw messages
+/// on codegen. Should never be used by application code.
+#[doc(hidden)]
+pub unsafe trait UpbGetArena: SealedInternal {
+    fn get_arena(&mut self, _private: Private) -> &Arena;
+}
+
+impl<T> MatcherEq for T
+where
+    Self: AssociatedMiniTable + AsView + Debug,
+    for<'a> View<'a, <Self as AsView>::Proxied>: UpbGetMessagePtr,
+{
+    fn matches(&self, o: &Self) -> bool {
+        unsafe {
+            upb_Message_IsEqual(
+                self.as_view().get_raw_message(Private),
+                o.as_view().get_raw_message(Private),
+                Self::mini_table(),
+                0,
+            )
+        }
+    }
+}
+
+impl<T: UpbGetMessagePtrMut> Clear for T {
+    fn clear(&mut self) {
+        unsafe { self.get_ptr_mut(Private).clear() }
+    }
+}
+
+fn clear_and_parse_helper<T>(
+    msg: &mut T,
+    data: &[u8],
+    decode_options: i32,
+) -> Result<(), ParseError>
+where
+    T: AssociatedMiniTable + UpbGetMessagePtrMut + UpbGetArena,
+{
+    Clear::clear(msg);
+    // SAFETY:
+    // - `msg` is a valid mutable message.
+    // - `mini_table` is the one associated with `msg`
+    // - `msg.arena().raw()` is held for the same lifetime as `msg`.
+    unsafe {
+        upb::wire::decode_with_options(
+            data,
+            msg.get_ptr_mut(Private),
+            msg.get_arena(Private),
+            decode_options,
+        )
+    }
+    .map(|_| ())
+    .map_err(|_| ParseError)
+}
+
+impl<T> ClearAndParse for T
+where
+    Self: AssociatedMiniTable + UpbGetMessagePtrMut + UpbGetArena,
+{
+    fn clear_and_parse(&mut self, data: &[u8]) -> Result<(), ParseError> {
+        clear_and_parse_helper(self, data, upb::wire::decode_options::CHECK_REQUIRED)
+    }
+
+    fn clear_and_parse_dont_enforce_required(&mut self, data: &[u8]) -> Result<(), ParseError> {
+        clear_and_parse_helper(self, data, 0)
+    }
+}
+
+impl<T> TakeFrom for T
+where
+    Self: CopyFrom + AsMut,
+    for<'a> Mut<'a, <Self as AsMut>::MutProxied>: Clear,
+{
+    fn take_from(&mut self, mut src: impl AsMut<MutProxied = Self::Proxied>) {
+        let mut src = src.as_mut();
+        // TODO: b/393559271 - Optimize this copy out.
+        CopyFrom::copy_from(self, AsView::as_view(&src));
+        Clear::clear(&mut src);
+    }
+}
+
+impl<T> CopyFrom for T
+where
+    Self: AsView + AssociatedMiniTable + UpbGetArena + UpbGetMessagePtr,
+    for<'a> View<'a, Self::Proxied>: UpbGetMessagePtr,
+{
+    fn copy_from(&mut self, src: impl AsView<Proxied = Self::Proxied>) {
+        // SAFETY: self and src are both valid `T`s associated with
+        // `Self::mini_table()`.
+        unsafe {
+            assert!(upb_Message_DeepCopy(
+                self.get_raw_message(Private),
+                src.as_view().get_raw_message(Private),
+                <Self as AssociatedMiniTable>::mini_table(),
+                self.get_arena(Private).raw()
+            ));
+        }
+    }
+}
+
+impl<T> MergeFrom for T
+where
+    Self: AsView + AssociatedMiniTable + UpbGetArena + UpbGetMessagePtr,
+    for<'a> View<'a, Self::Proxied>: UpbGetMessagePtr,
+{
+    fn merge_from(&mut self, src: impl AsView<Proxied = Self::Proxied>) {
+        // SAFETY: self and src are both valid `T`s.
+        unsafe {
+            assert!(upb_Message_MergeFrom(
+                self.get_raw_message(Private),
+                src.as_view().get_raw_message(Private),
+                <Self as AssociatedMiniTable>::mini_table(),
+                // Use a nullptr for the ExtensionRegistry.
+                std::ptr::null(),
+                self.get_arena(Private).raw()
+            ));
+        }
+    }
+}
